"""This module contains functions to postprocess and serialize results."""

import logging
import tempfile
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import pandas as pd

from scythe.types import S3Url

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = object

logger = logging.getLogger(__name__)


def serialize_df_dict(dfs: dict[str, pd.DataFrame]) -> dict[str, dict]:
    """Serialize a dictionary of dataframes into a dictionary of dictionaries.

    Args:
        dfs (dict[str, pd.DataFrame]): A dictionary of dataframes

    Returns:
        dict[str, dict]: A dictionary of dictionaries
    """
    return {k: v.to_dict(orient="tight") for k, v in dfs.items()}


def transpose_dataframe_dict(
    dataframe_results: list[dict[str, pd.DataFrame]],
) -> dict[str, pd.DataFrame]:
    """Transpose a list of dictionaries of dataframes into a dictionary of combined dataframes."""
    all_keys = {key for df_dict in dataframe_results for key in df_dict}
    return {
        key: pd.concat(
            [df_dict[key] for df_dict in dataframe_results if key in df_dict], axis=0
        )
        for key in all_keys
    }


def make_onerow_multiindex_from_dict(
    d: dict[str, Any], n_rows: int = 1
) -> pd.MultiIndex:
    """Makes a MultiIndex from a dictionary.

    This is useful for returning a wide-form dataframe of results for a single task.

    Args:
        d (dict[str, Any]): The dictionary to make the MultiIndex from.
        n_rows (int): The number of rows to repeat the MultiIndex.

    Returns:
        multi_index (pd.MultiIndex): The MultiIndex.
    """
    return pd.MultiIndex.from_tuples(
        [tuple(d.values())] * n_rows,
        names=list(d.keys()),
    )


def save_and_upload_parquets(
    collected_dfs: dict[str, pd.DataFrame],
    s3: S3Client,
    bucket: str,
    output_key_constructor: Callable[[str], str],
    save_errors: bool = False,
) -> dict[str, S3Url]:
    """Save and upload results to s3."""
    uris: dict[str, S3Url] = {}
    with tempfile.TemporaryDirectory() as tmpdir:
        for key, df in collected_dfs.items():
            output_key = output_key_constructor(key)
            if "error" in key.lower() and not save_errors:
                continue
            f = f"{tmpdir}/{key}.parquet"
            df.to_parquet(f)
            s3.upload_file(Bucket=bucket, Key=output_key, Filename=f)
            uri = f"s3://{bucket}/{output_key}"
            uris[key] = S3Url(uri)
    return uris



"""Fanout Handling."""

import tempfile
from functools import cached_property, partial
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeVar

import boto3
import pandas as pd
from hatchet_sdk import Context, TriggerWorkflowOptions
from hatchet_sdk.clients.admin import WorkflowRunTriggerConfig
from pydantic import BaseModel, Field, field_validator, model_validator

from scythe.base import BaseSpec, ExperimentInputSpec, ExperimentOutputSpec
from scythe.hatchet import hatchet
from scythe.registry import ExperimentRegistry, ExperimentStandaloneType, TOutput
from scythe.storage import ScytheStorageSettings
from scythe.types import S3Url
from scythe.utils.filesys import fetch_uri
from scythe.utils.results import save_and_upload_parquets, transpose_dataframe_dict

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = object


s3 = boto3.client("s3")


class RecursionSpec(BaseModel):
    """A spec for recursive calls."""

    factor: int = Field(..., description="The factor to use in recursive calls", ge=1)
    offset: int | None = Field(
        default=None, description="The offset to use in recursive calls", ge=0
    )

    @model_validator(mode="before")
    @classmethod
    def validate_offset_less_than_factor(cls, values):
        """Validate that the offset is less than the factor."""
        if values["offset"] is None:
            return values
        if values["offset"] >= values["factor"]:
            raise ValueError(f"OFFSET:{values['offset']}>=FACTOR:{values['factor']}")
        return values


class RecursionMap(BaseModel):
    """A map of recursion specs to use in recursive calls.

    This allows a recursion node to understand where
    it is in the recursion tree and how to behave.
    """

    path: list[RecursionSpec] | None = Field(
        default=None, description="The path of recursion specs to use"
    )
    factor: int = Field(..., description="The factor to use in recursive calls", ge=1)
    max_depth: int = Field(
        default=10, description="The maximum depth of the recursion", ge=0, le=10
    )

    @property
    def is_root(self) -> bool:
        """Check if the recursion map is the root."""
        return len(self.path or []) == 0

    @field_validator("path", mode="before")
    @classmethod
    def validate_path_is_length_ge_1(cls, values):
        """Validate that the path is at least length 1."""
        if values is None:
            return values
        if len(values) < 1:
            raise ValueError("PATH:LENGTH:GE:1")
        return values


class GatheredExperimentRuns(BaseModel, arbitrary_types_allowed=True):
    """A class for gathering experiment results."""

    # TODO: make success a little more extensible?
    success: ExperimentOutputSpec
    errors: pd.DataFrame | None


class ScatterGatherInput(BaseSpec):
    """Input for the scatter gather workflow."""

    task_name: str = Field(..., description="The name of the Hatchet task to run.")
    # TODO: consider isolating specs path to allow two versions of the class, including
    # one which takes specs rather than S3Url
    specs_uri: S3Url = Field(..., description="The path to the specs to run.")
    storage_settings: ScytheStorageSettings = Field(
        default_factory=lambda: ScytheStorageSettings(),
        description="The storage settings to use.",
    )
    recursion_map: RecursionMap = Field(..., description="The recursion map to use.")
    save_errors: bool = Field(default=False, description="Whether to save errors.")

    def construct_filekey(
        self,
        filename: str,
        *,
        mode: Literal["input", "output", "final"],
        workflow_run_id: str,
        suffix: str,
    ) -> str:
        """Cosntruct an output key for the scatter gather workflow."""
        bucket_prefix = self.storage_settings.BUCKET_PREFIX
        output_key_base = f"{bucket_prefix}/{self.experiment_id}/"
        mode_prefix = f"{mode}/" if mode == "final" else f"scatter-gather/{mode}/"
        recurse_prefix = (
            (
                f"recurse/{len(self.recursion_map.path)}/"
                if self.recursion_map.path
                else "root/"
            )
            if mode != "final"
            else ""
        )
        workflow_run_id_prefix = f"{workflow_run_id}/" if mode != "final" else ""
        output_key_prefix = (
            f"{output_key_base}{mode_prefix}{recurse_prefix}{workflow_run_id_prefix}"
        )
        return f"{output_key_prefix}/{filename}.{suffix}"

    @property
    def standalone(self) -> ExperimentStandaloneType:
        """Get the experiment standalone."""
        return ExperimentRegistry.get_experiment(self.task_name)

    @cached_property
    def specs(self) -> list[ExperimentInputSpec]:
        """Fetch the specs and convert to the input type."""
        local_path = self.fetch_uri(self.specs_uri, use_cache=True)
        df = pd.read_parquet(local_path)
        specs_dicts = df.to_dict(orient="records")
        validator = self.standalone.input_validator
        return [validator.model_validate(spec) for spec in specs_dicts]

    def add_root_workflow_run_id(self, root_workflow_run_id: str) -> None:
        """Add the root workflow run id to the specs."""
        if self.recursion_map.is_root:
            for spec in self.specs:
                spec.root_workflow_run_id = root_workflow_run_id

    async def run_experiments(
        self,
    ) -> GatheredExperimentRuns:
        """Run the actual experiments and collect results."""
        specs = self.specs
        inputs = [
            self.standalone.create_bulk_run_item(
                input=spec,
                options=TriggerWorkflowOptions(
                    additional_metadata={"index": i},
                ),
            )
            for i, spec in enumerate(specs)
        ]
        # since we don't have a way to get the individual workflow run ids,
        # we can't associate specific results (or errors) with specific results
        results = await self.standalone.aio_run_many(inputs, return_exceptions=True)
        safe_results, error_df = sift_results(specs, results)
        result = combine_experiment_outputs(safe_results)
        return GatheredExperimentRuns(
            success=result,
            errors=error_df,
        )

    def create_recursion_payloads(
        self, parent_workflow_run_id: str
    ) -> tuple[list[WorkflowRunTriggerConfig], list["ScatterGatherInput"]]:
        """Splits up the specs into a list of children scatter gather payloads for recursion."""
        specs = self.specs
        children_payloads: list[ScatterGatherInput] = []
        for branch_ix in range(self.recursion_map.factor):
            new_path = self.recursion_map.path.copy() if self.recursion_map.path else []
            new_path.append(
                RecursionSpec(factor=self.recursion_map.factor, offset=branch_ix)
            )
            recursion_map = RecursionMap(
                path=new_path,
                factor=self.recursion_map.factor,
                max_depth=self.recursion_map.max_depth,
            )
            specs_to_use = specs[branch_ix :: self.recursion_map.factor]
            specs_as_df = pd.DataFrame([
                spec.model_dump(mode="json") for spec in specs_to_use
            ])
            filename = f"specs_{branch_ix:06d}"
            uris = save_and_upload_parquets(
                collected_dfs={
                    filename: specs_as_df,
                },
                s3=s3,
                bucket=self.storage_settings.BUCKET,
                output_key_constructor=partial(
                    self.construct_filekey,
                    mode="input",
                    workflow_run_id=parent_workflow_run_id,
                    suffix="pq",
                ),
            )
            payload = ScatterGatherInput(
                experiment_id=self.experiment_id,
                task_name=self.task_name,
                storage_settings=self.storage_settings,
                specs_uri=uris[filename],
                recursion_map=recursion_map,
            )
            del specs_as_df
            del specs_to_use
            children_payloads.append(payload)
        triggers = [
            scatter_gather.create_bulk_run_item(
                input=payload,
                options=TriggerWorkflowOptions(
                    additional_metadata={
                        "index": i,
                        "level": len(self.recursion_map.path or []),
                        "experiment_id": self.experiment_id,
                        "experiment_name": self.task_name,
                    },
                ),
            )
            for i, payload in enumerate(children_payloads)
        ]
        return triggers, children_payloads

    @property
    def is_root(self) -> bool:
        """Check if the current payload is a root, i.e. the original call."""
        return self.recursion_map.is_root

    @property
    def is_base_case(self) -> bool:
        """Check if the current payload is a base case, i.e. no recursion needed."""
        too_few_specs = len(self.specs) <= self.recursion_map.factor
        past_max_depth = (
            (len(self.recursion_map.path) >= self.recursion_map.max_depth)
            if self.recursion_map.path
            else False
        ) or self.recursion_map.max_depth == 0
        return too_few_specs or past_max_depth


class ScatterGatherResult(BaseModel):
    """The result of the scatter gather workflow."""

    uris: dict[str, S3Url]

    def to_gathered_experiment_runs(self) -> GatheredExperimentRuns:
        """Convert the scatter gather result to a gathered experiment runs."""
        dfs: dict[str, pd.DataFrame] = {}
        for k, v in self.uris.items():
            with tempfile.TemporaryDirectory() as tmpdir:
                f = Path(tmpdir) / "specs.parquet"
                res_path = fetch_uri(uri=v, local_path=f, use_cache=False)
                dfs[k] = pd.read_parquet(res_path)
        return GatheredExperimentRuns(
            success=ExperimentOutputSpec(
                dataframes=dfs,
            ),
            errors=None,
        )


@hatchet.task(name="scatter_gather", input_validator=ScatterGatherInput)
async def scatter_gather(
    payload: ScatterGatherInput,
    ctx: Context,
) -> ScatterGatherResult:
    """Run the scatter gather workflow."""
    payload.add_root_workflow_run_id(ctx.workflow_run_id)
    if payload.is_base_case:
        experiment_output = await payload.run_experiments()
        experiment_outputs = [experiment_output]
    else:
        trigs, children_specs = payload.create_recursion_payloads(ctx.workflow_run_id)
        results = await scatter_gather.aio_run_many(trigs, return_exceptions=True)
        safe_results, error_df = sift_results(children_specs, results)
        experiment_outputs = [r.to_gathered_experiment_runs() for r in safe_results]

    dfs = [r.success.dataframes for r in experiment_outputs]
    errors = [r.errors for r in experiment_outputs if r.errors is not None]
    transposed_dfs = transpose_dataframe_dict(dfs)
    if errors:
        error_dfs = pd.concat(errors, axis=0)
        if "errors" in transposed_dfs:
            transposed_dfs["errors"] = pd.concat(
                [transposed_dfs["errors"], error_dfs], axis=0
            )
        else:
            transposed_dfs["errors"] = error_dfs
    output_key_constructor = partial(
        payload.construct_filekey,
        mode="output",
        workflow_run_id=ctx.workflow_run_id,
        suffix="pq",
    )
    uris = save_and_upload_parquets(
        collected_dfs=transposed_dfs,
        s3=s3,
        bucket=payload.storage_settings.BUCKET,
        output_key_constructor=output_key_constructor,
        save_errors=payload.save_errors,
    )
    if payload.is_root:
        output_key_constructor = partial(
            payload.construct_filekey,
            mode="final",
            workflow_run_id=ctx.workflow_run_id,
            suffix="pq",
        )
        uris = save_and_upload_parquets(
            collected_dfs=transposed_dfs,
            s3=s3,
            bucket=payload.storage_settings.BUCKET,
            output_key_constructor=output_key_constructor,
            save_errors=payload.save_errors,
        )

    return ScatterGatherResult(uris=uris)


ZipDataContent = TypeVar("ZipDataContent", bound=BaseModel)
ResultDataContent = TypeVar("ResultDataContent")


def sift_results(
    spec_data: list[ZipDataContent],
    results: list[ResultDataContent | BaseException],
) -> tuple[list[ResultDataContent], pd.DataFrame | None]:
    """Sift results into safe and errored results.

    Args:
        spec_data (list[ZipDataContent]): The list of spec data.
        results (list[ResultDataContent | BaseException]): The list of results.

    Returns:
        safe_results (list[ResultDataContent]): The list of safe results.
        errors (pd.DataFrame): The DataFrame of errored results.
    """
    safe_results = [r for r in results if not isinstance(r, BaseException)]
    error_specs = [
        spec_data[i] for i, r in enumerate(results) if isinstance(r, BaseException)
    ]

    if error_specs:
        error_msgs = [str(r) for r in results if isinstance(r, BaseException)]
        error_multi_index = pd.MultiIndex.from_frame(
            pd.DataFrame([r.model_dump(mode="json") for r in error_specs])
        )
        error_df = pd.DataFrame(
            {"missing": [False], "msg": error_msgs}, index=error_multi_index
        )
    else:
        error_df = None
    return safe_results, error_df


def combine_experiment_outputs(
    experiment_outputs: list[TOutput],
) -> ExperimentOutputSpec:
    """Combine a list of experiment outputs into a single experiment output."""
    # TODO: handle other keys besides dataframes.
    dataframes = transpose_dataframe_dict([
        output.dataframes for output in experiment_outputs
    ])
    return ExperimentOutputSpec(
        dataframes=dataframes,
    )


if __name__ == "__main__":
    import json

    dataframes = {
        "results": pd.DataFrame({"a": [1, 2, 3]}),
        "metrics": pd.DataFrame({"b": [4, 5, 6]}),
    }
    experiment_output = ExperimentOutputSpec(dataframes=dataframes)
    print(experiment_output.model_dump())
    print(
        experiment_output.model_validate(
            json.loads(experiment_output.model_dump_json())
        )
    )



"""Storage Settings."""

from pydantic import Field
from pydantic_settings import BaseSettings


class ScytheStorageSettings(BaseSettings, env_prefix="SCYTHE_STORAGE_"):
    """Storage Settings."""

    BUCKET: str = Field(default=..., description="The name of the S3 bucket to use.")
    BUCKET_PREFIX: str = Field(
        default="scythe", description="The prefix to use for the S3 bucket."
    )



"""Worker for Scythe."""

import os

from pydantic import Field
from pydantic_settings import BaseSettings

from scythe.hatchet import hatchet
from scythe.registry import ExperimentFunction, ExperimentRegistry
from scythe.scatter_gather import scatter_gather


class WorkerNameConfig(BaseSettings):
    """Configuration for the worker name."""

    FLY_REGION: str | None = None
    AWS_BATCH_JOB_ARRAY_INDEX: int | None = None
    COPILOT_ENVIRONMENT_NAME: str | None = None

    @property
    def in_aws_batch(self) -> bool:
        """Return whether the worker is running in AWS Batch."""
        return self.AWS_BATCH_JOB_ARRAY_INDEX is not None

    @property
    def in_aws_copilot(self) -> bool:
        """Return whether the worker is running in AWS Copilot."""
        return self.COPILOT_ENVIRONMENT_NAME is not None

    @property
    def in_aws(self) -> bool:
        """Return whether the worker is running in AWS."""
        return self.in_aws_batch or self.in_aws_copilot

    @property
    def aws_hosting_str(self):
        """Return the AWS hosting string for the worker."""
        batch = (
            f"Batch{self.AWS_BATCH_JOB_ARRAY_INDEX:04d}" if self.in_aws_batch else ""
        )
        copilot = (
            f"Copilot{(self.COPILOT_ENVIRONMENT_NAME or '').upper()}"
            if self.in_aws_copilot
            else ""
        )
        return f"AWS{batch}{copilot}" if self.in_aws else ""

    @property
    def in_fly(self) -> bool:
        """Return whether the worker is running in Fly.io."""
        return self.FLY_REGION is not None

    @property
    def fly_hosting_str(self):
        """Return the Fly hosting string for the worker."""
        return f"Fly{(self.FLY_REGION or '').upper()}" if self.in_fly else ""

    @property
    def in_local(self) -> bool:
        """Return whether the worker is running locally."""
        return not self.in_aws and not self.in_fly

    @property
    def hosting_str(self):
        """Return the hosting string for the worker."""
        return self.aws_hosting_str or self.fly_hosting_str or "Local"

    @property
    def name(self) -> str:
        """Return the name of the worker."""
        base = "ScytheWorker"
        return f"{base}--{self.hosting_str}"


class ScytheWorkerConfig(BaseSettings, env_prefix="SCYTHE_WORKER_"):
    """Configuration for the Scythe worker."""

    NAME: str | None = Field(default=None, description="The name of the worker.")
    SLOTS: int | None = Field(
        default=None, description="The number of slots the worker has."
    )
    DURABLE_SLOTS: int | None = Field(
        default=None, description="The number of durable slots the worker has."
    )
    HIGH_MEMORY: bool = Field(
        default=False, description="Whether the worker has high memory."
    )
    HIGH_CPU: bool = Field(
        default=False, description="Whether the worker has high CPU."
    )

    WORKER_NAME_CONFIG: WorkerNameConfig = Field(
        default_factory=WorkerNameConfig,
        description="The configuration for the worker name.",
    )

    DOES_FAN: bool = True
    DOES_LEAF: bool = True

    @property
    def labels(self) -> dict[str, str | int]:
        """Return the labels for the worker."""
        return {
            "high_memory": self.HIGH_MEMORY,
            "high_cpu": self.HIGH_CPU,
        }

    @property
    def computed_slots(self) -> int:
        """Return the number of slots for the worker."""
        if self.SLOTS is not None:
            return self.SLOTS
        cpu_ct = os.cpu_count() or 1
        if cpu_ct < 8:
            return cpu_ct
        else:
            return cpu_ct - 1

    @property
    def computed_durable_slots(self) -> int:
        """Return the number of durable slots for the worker."""
        if self.DURABLE_SLOTS is not None:
            return self.DURABLE_SLOTS
        return 1000

    @property
    def computed_name(self) -> str:
        """Return the name of the worker."""
        return self.NAME if self.NAME else self.WORKER_NAME_CONFIG.name

    def start(self, experiments: list[ExperimentFunction] | None = None) -> None:
        """Make a worker."""
        for experiment in experiments or []:
            ExperimentRegistry.Register()(experiment)

        worker = hatchet.worker(
            name=self.computed_name,
            slots=self.computed_slots,
            durable_slots=self.computed_durable_slots,
            labels=self.labels,
        )
        workflows = ([scatter_gather] if self.DOES_FAN else []) + (
            ExperimentRegistry.experiments() if self.DOES_LEAF else []
        )
        for workflow in workflows:
            worker.register_workflow(workflow)
        worker.start()



"""Types for Scythe."""

from pathlib import Path

from pydantic import AnyUrl, FilePath, HttpUrl, UrlConstraints


class S3Url(AnyUrl):
    """A URL for an S3 object."""

    _constraints = UrlConstraints(allowed_schemes=["s3"], host_required=True)


FileReference = S3Url | HttpUrl | Path | FilePath



"""Register experiments with Scythe."""

from datetime import timedelta
from typing import ClassVar, Protocol, TypeVar, cast, get_type_hints

from hatchet_sdk import Context, Worker
from hatchet_sdk.labels import DesiredWorkerLabel
from hatchet_sdk.runnables.workflow import Standalone
from hatchet_sdk.utils.timedelta_to_expression import Duration

from scythe.base import ExperimentInputSpec, ExperimentOutputSpec
from scythe.hatchet import hatchet

ExperimentStandaloneType = Standalone[ExperimentInputSpec, ExperimentOutputSpec]

TInput = TypeVar("TInput", bound=ExperimentInputSpec, contravariant=True)
TOutput = TypeVar("TOutput", bound=ExperimentOutputSpec, covariant=True)


class ExperimentTypeNotFound(Exception):
    """An experiment was not found."""

    def __init__(self, name: str):
        """Initialize the exception."""
        self.name = name
        super().__init__(f"Experiment {name} not found")


class ExperimentTypeExists(Exception):
    """An experiment type already exists."""

    def __init__(self, name: str):
        """Initialize the exception."""
        self.name = name
        super().__init__(f"Experiment {name} already exists")


class ExperimentFunction(Protocol[TInput, TOutput]):
    """A function that takes an input spec and returns an output spec."""

    def __call__(self, input_spec: TInput) -> TOutput:
        """Invoke the experiment function.

        Args:
            input_spec: The input spec for the experiment.
        """
        ...

    def __name__(self) -> str:
        """The name of the experiment function."""
        ...


class ExperimentRegistry:
    """An experiment registry for standalone task steps."""

    _experiments_dict: ClassVar[dict[str, ExperimentStandaloneType]] = {}

    @classmethod
    def Include(
        cls,
        task: Standalone[TInput, TOutput],
    ) -> ExperimentStandaloneType:
        """Register an experiment whose input and output schemas conform to the base types."""
        task_safe = cast(ExperimentStandaloneType, task)
        if task_safe.name in cls._experiments_dict:
            raise ExperimentTypeExists(task_safe.name)
        cls._experiments_dict[task_safe.name] = task_safe
        return task_safe

    @classmethod
    def Register(
        cls,
        *,
        worker: Worker | None = None,
        description: str | None = None,
        name: str | None = None,
        desired_worker_labels: dict[str, DesiredWorkerLabel] | None = None,
        schedule_timeout: Duration = timedelta(minutes=5),
        execution_timeout: Duration = timedelta(minutes=1),
        retries: int = 1,
        overwrite_log_method: bool = True,
        inject_workflow_run_id: bool = True,
        **task_config,
    ):
        """Decorator to make a standalone experiment from a function.

        Usage:
            @ExperimentRegistry.Register(description="desc", ...)
            def my_experiment(input_spec: MyInput) -> MyOutput:
                ...
        """

        def decorator(
            fn: ExperimentFunction[TInput, TOutput],
        ) -> ExperimentStandaloneType:
            input_type = cast(type[TInput], get_type_hints(fn)["input_spec"])
            return_type = cast(type[TOutput], get_type_hints(fn)["return"])

            fn_name = fn.__name__
            fn_doc = fn.__doc__

            @cls.Include
            @hatchet.task(
                name=name or f"scythe_experiment_{fn_name}",
                description=description or f"{fn_doc}",
                input_validator=input_type,
                desired_worker_labels=desired_worker_labels,
                schedule_timeout=schedule_timeout,
                execution_timeout=execution_timeout,
                retries=retries,
                **task_config,
            )
            def task(input_: input_type, context: Context) -> return_type:  # pyright: ignore [reportInvalidTypeForm]
                """The task implementation."""
                if overwrite_log_method:
                    input_.log = lambda msg: context.log(msg)
                if inject_workflow_run_id:
                    input_.workflow_run_id = context.workflow_run_id
                output = fn(input_)
                output.add_scalars(input_)
                return output

            if worker:
                worker.register_workflow(task)
            return task

        return decorator

    @classmethod
    def get_experiment(cls, name: str) -> ExperimentStandaloneType:
        """Get an experiment's Hatchet Stanadalone."""
        if name not in cls._experiments_dict:
            raise ExperimentTypeNotFound(name)
        return cls._experiments_dict[name]

    @classmethod
    def experiments(cls) -> list[ExperimentStandaloneType]:
        """Get all experiments."""
        return list(cls._experiments_dict.values())


# if __name__ == "__main__":
